"""This module was copied from here in order to remove the mopidy dependency in the server code:
https://github.com/mopidy/mopidy-spotify/blob/master/mopidy_spotify/web.py"""
# pylint: skip-file
import copy
import email
import logging
import os
import re
import time
import urllib.parse
from dataclasses import dataclass
from datetime import datetime
from enum import Enum, unique
from typing import Optional

import requests


logger = logging.getLogger(__name__)


def get_requests_session(proxy_config):
    """This method was pulled from
    https://github.com/mopidy/mopidy-spotify/blob/master/mopidy_spotify/utils.py
    The possibility for proxy configuration was dropped and a static user agent is used
    in order to remove the dependency to mopdiy."""
    full_user_agent = "Mopidy-Spotify/4.0.1 Mopidy/3.0.2 CPython/3.7.5"

    session = requests.Session()
    session.headers.update({"user-agent": full_user_agent})

    return session


def _trace(*args, **kwargs):
    """This line was removed to remove the utils dependency."""
    # logger.log(utils.TRACE, *args, **kwargs)
    pass


class OAuthTokenRefreshError(Exception):
    def __init__(self, reason):
        message = f"OAuth token refresh failed: {reason}"
        super().__init__(message)


class OAuthClientError(Exception):
    pass


class OAuthClient:
    def __init__(
        self,
        base_url,
        refresh_url,
        client_id=None,
        client_secret=None,
        proxy_config=None,
        expiry_margin=60,
        timeout=10,
        retries=3,
        retry_statuses=(500, 502, 503, 429),
    ):

        if client_id and client_secret:
            self._auth = (client_id, client_secret)
        else:
            self._auth = None

        self._base_url = base_url
        self._refresh_url = refresh_url

        self._margin = expiry_margin
        self._expires = 0
        self._authorization_failed = False

        self._timeout = timeout
        self._number_of_retries = retries
        self._retry_statuses = retry_statuses
        self._backoff_factor = 0.5

        self._headers = {"Content-Type": "application/json"}
        self._session = get_requests_session(proxy_config or {})

    def get(self, path, cache=None, *args, **kwargs):
        if self._authorization_failed:
            logger.debug("Blocking request as previous authorization failed.")
            return WebResponse(None, None)

        params = kwargs.pop("params", None)
        path = self._normalise_query_string(path, params)

        _trace(f"Get '{path}'")

        ignore_expiry = kwargs.pop("ignore_expiry", False)
        if cache is not None and path in cache:
            cached_result = cache.get(path)
            if cached_result.still_valid(ignore_expiry):
                return cached_result
            kwargs.setdefault("headers", {}).update(cached_result.etag_headers)

        # TODO: Factor this out once we add more methods.
        # TODO: Don't silently error out.
        try:
            if self._should_refresh_token():
                self._refresh_token()
        except OAuthTokenRefreshError as e:
            logger.error(e)
            return WebResponse(None, None)

        # Make sure our headers always override user supplied ones.
        kwargs.setdefault("headers", {}).update(self._headers)
        result = self._request_with_retries("GET", path, *args, **kwargs)

        if result is None or "error" in result:
            logger.error(
                "Spotify Web API request failed: " f"{result.get('error','Unknown')}"
            )
            return WebResponse(None, None)

        if self._should_cache_response(cache, result):
            previous_result = cache.get(path)
            if previous_result and previous_result.updated(result):
                result = previous_result
            cache[path] = result

        return result

    def _should_cache_response(self, cache, response):
        return cache is not None and response.status_ok

    def _should_refresh_token(self):
        # TODO: Add jitter to margin?
        return not self._auth or time.time() > self._expires - self._margin

    def _refresh_token(self):
        logger.debug(f"Fetching OAuth token from {self._refresh_url}")

        data = {"grant_type": "client_credentials"}
        result = self._request_with_retries(
            "POST", self._refresh_url, auth=self._auth, data=data
        )

        if result is None:
            raise OAuthTokenRefreshError("Unknown error.")
        elif result.get("error"):
            raise OAuthTokenRefreshError(
                f"{result['error']} {result.get('error_description', '')}"
            )
        elif not result.get("access_token"):
            raise OAuthTokenRefreshError("missing access_token")
        elif result.get("token_type") != "Bearer":
            raise OAuthTokenRefreshError(
                f"wrong token_type: {result.get('token_type')}"
            )

        self._headers["Authorization"] = f"Bearer {result['access_token']}"
        self._expires = time.time() + result.get("expires_in", float("Inf"))

        if result.get("expires_in"):
            logger.debug(f"Token expires in {result['expires_in']} seconds.",)
        if result.get("scope"):
            logger.debug(f"Token scopes: {result['scope']}")

    def _request_with_retries(self, method, url, *args, **kwargs):
        prepared_request = self._session.prepare_request(
            requests.Request(method, self._prepare_url(url, *args), **kwargs)
        )

        try_until = time.time() + self._timeout

        result = None
        backoff_time = 0

        for i in range(self._number_of_retries):
            remaining_timeout = max(try_until - time.time(), 1)

            # Give up if we don't have any timeout left after sleeping.
            if backoff_time > remaining_timeout:
                break
            elif backoff_time > 0:
                time.sleep(backoff_time)

            try:
                response = self._session.send(
                    prepared_request, timeout=remaining_timeout
                )
            except requests.RequestException as e:
                logger.debug(f"Fetching {prepared_request.url} failed: {e}")
                status_code = None
                backoff_time = 0
                result = None
            else:
                status_code = response.status_code
                backoff_time = self._parse_retry_after(response)
                result = WebResponse.from_requests(prepared_request, response)

            if status_code and 400 <= status_code < 600:
                logger.debug(f"Fetching {prepared_request.url} failed: {status_code}")

            # Filter out cases where we should not retry.
            if status_code and status_code not in self._retry_statuses:
                break

            # TODO: Provider might return invalid JSON for "OK" responses.
            # This should really not happen, so ignoring for the purpose of
            # retries. It would be easier if they correctly used 204, but
            # instead some endpoints return 200 with no content, or true/false.

            # Decide how long to sleep in the next iteration.
            backoff_time = backoff_time or (2 ** i * self._backoff_factor)
            logger.debug(
                f"Retrying {prepared_request.url} in {backoff_time:.3f} " "seconds."
            )

        if status_code == 401:
            self._authorization_failed = True
            logger.error(
                "Authorization failed, not attempting Spotify API "
                "request. Please get new credentials from "
                "https://www.mopidy.com/authenticate and/or restart "
                "Mopidy to resolve this problem."
            )
        return result

    def _prepare_url(self, url, *args, **kwargs):
        # TODO: Move this out as a helper and unit-test it directly?
        b = urllib.parse.urlsplit(self._base_url)
        u = urllib.parse.urlsplit(url.format(*args))

        if u.scheme or u.netloc:
            scheme, netloc, path = u.scheme, u.netloc, u.path
            query = urllib.parse.parse_qsl(u.query, keep_blank_values=True)
        else:
            scheme, netloc = b.scheme, b.netloc
            path = os.path.normpath(os.path.join(b.path, u.path))
            query = urllib.parse.parse_qsl(b.query, keep_blank_values=True)
            query.extend(urllib.parse.parse_qsl(u.query, keep_blank_values=True))

        for key, value in kwargs.items():
            query.append((key, value))

        encoded_query = urllib.parse.urlencode(dict(query))
        return urllib.parse.urlunsplit((scheme, netloc, path, encoded_query, ""))

    def _normalise_query_string(self, url, params=None):
        u = urllib.parse.urlsplit(url)
        scheme, netloc, path = u.scheme, u.netloc, u.path

        query = dict(urllib.parse.parse_qsl(u.query, keep_blank_values=True))
        if isinstance(params, dict):
            query.update(params)
        sorted_unique_query = sorted(query.items())
        encoded_query = urllib.parse.urlencode(sorted_unique_query)
        return urllib.parse.urlunsplit((scheme, netloc, path, encoded_query, ""))

    def _parse_retry_after(self, response):
        """Parse Retry-After header from response if it is set."""
        value = response.headers.get("Retry-After")

        if not value:
            seconds = 0
        elif re.match(r"^\s*[0-9]+\s*$", value):
            seconds = int(value)
        else:
            date_tuple = email.utils.parsedate(value)
            if date_tuple is None:
                seconds = 0
            else:
                seconds = time.mktime(date_tuple) - time.time()
        return max(0, seconds)


class WebResponse(dict):
    def __init__(self, url, data, expires=0.0, etag=None, status_code=400):
        self._from_cache = False
        self.url = url
        self._expires = expires
        self._etag = etag
        self._status_code = status_code
        super().__init__(data or {})
        _trace(f"New WebResponse {self}")

    @classmethod
    def from_requests(cls, request, response):
        expires = cls._parse_cache_control(response)
        etag = cls._parse_etag(response)
        json = cls._decode(response)
        return cls(request.url, json, expires, etag, response.status_code)

    @staticmethod
    def _decode(response):
        # Deal with 204 and other responses with empty body.
        if not response.content:
            return None
        try:
            return response.json()
        except ValueError as e:
            url = response.request.url
            logger.error(f"JSON decoding {url} failed: {e}")
            return None

    @staticmethod
    def _parse_cache_control(response):
        """Parse Cache-Control header from response if it is set."""
        value = response.headers.get("Cache-Control", "no-store").lower()

        if "no-store" in value:
            seconds = 0
        else:
            max_age = re.match(r".*max-age=\s*([0-9]+)\s*", value)
            if not max_age:
                seconds = 0
            else:
                seconds = int(max_age.groups()[0])
        return time.time() + seconds

    @staticmethod
    def _parse_etag(response):
        """Parse ETag header from response if it is set."""
        value = response.headers.get("ETag")

        if value:
            # 'W/' (case-sensitive) indicates that a weak validator is used,
            # currently ignoring this.
            # Format is string of ASCII characters placed between double quotes
            # but can seemingly also include hyphen characters.
            etag = re.match(r'^(W/)?("[!#-~]+")$', value)
            if etag and len(etag.groups()) == 2:
                return etag.groups()[1]

    def still_valid(self, ignore_expiry=False):
        if ignore_expiry:
            result = True
            status = "forced"
        elif self._expires >= time.time():
            result = True
            status = "fresh"
        else:
            result = False
            status = "expired"
        self._from_cache = result
        _trace("Cached data %s for %s", status, self)
        return result

    @property
    def status_unchanged(self):
        return self._from_cache or 304 == self._status_code

    @property
    def status_ok(self):
        return self._status_code >= 200 and self._status_code < 400

    @property
    def etag_headers(self):
        if self._etag is not None:
            return {"If-None-Match": self._etag}
        else:
            return {}

    def updated(self, response):
        self._from_cache = False
        if self._etag is None:
            return False
        elif self.url != response.url:
            logger.error(f"ETag mismatch (different URI) for {self} {response}")
            return False
        elif not response.status_ok:
            logger.debug(f"ETag mismatch (bad response) for {self} {response}")
            return False
        elif response._status_code != 304:
            _trace(f"ETag mismatch for {self} {response}")
            return False

        _trace(f"ETag match for {self} {response}")
        self._expires = response._expires
        self._etag = response._etag
        self._status_code = response._status_code
        return True

    def __str__(self):
        return (
            f"URL: {self.url} "
            f"expires at: {datetime.fromtimestamp(self._expires)} "
            f"[ETag: {self._etag}]"
        )

    def increase_expiry(self, delta_seconds):
        if self.status_ok and not self._from_cache:
            self._expires += delta_seconds


class SpotifyOAuthClient(OAuthClient):

    TRACK_FIELDS = (
        "next,items(track(type,uri,name,duration_ms,disc_number,track_number,"
        "artists,album,is_playable,linked_from.uri))"
    )
    PLAYLIST_FIELDS = f"name,owner.id,type,uri,snapshot_id,tracks({TRACK_FIELDS}),"
    DEFAULT_EXTRA_EXPIRY = 10

    def __init__(self, *, client_id, client_secret, proxy_config):
        super().__init__(
            base_url="https://api.spotify.com/v1",
            refresh_url="https://auth.mopidy.com/spotify/token",
            client_id=client_id,
            client_secret=client_secret,
            proxy_config=proxy_config,
        )
        self.user_id = None
        self._cache = {}
        self._extra_expiry = self.DEFAULT_EXTRA_EXPIRY

    def get_one(self, path, *args, **kwargs):
        _trace(f"Fetching page {path!r}")
        result = self.get(path, cache=self._cache, *args, **kwargs)
        result.increase_expiry(self._extra_expiry)
        return result

    def get_all(self, path, *args, **kwargs):
        while path is not None:
            result = self.get_one(path, *args, **kwargs)
            path = result.get("next")
            yield result

    def login(self):
        self.user_id = self.get("me").get("id")
        if self.user_id is None:
            logger.error("Failed to load Spotify user profile")
            return False
        else:
            logger.info(f"Logged into Spotify Web API as {self.user_id}")
            return True

    @property
    def logged_in(self):
        return self.user_id is not None

    def get_user_playlists(self):
        pages = self.get_all("me/playlists", params={"limit": 50})
        for page in pages:
            yield from page.get("items", [])

    def get_playlist(self, uri):
        try:
            parsed = WebLink.from_uri(uri)
            if parsed.type != LinkType.PLAYLIST:
                raise ValueError(f"Could not parse {uri!r} as a Spotify playlist URI")
        except ValueError as exc:
            logger.error(exc)
            return {}

        playlist = self.get_one(
            f"playlists/{parsed.id}",
            params={"fields": self.PLAYLIST_FIELDS, "market": "from_token"},
        )

        tracks_path = playlist.get("tracks", {}).get("next")
        track_pages = self.get_all(
            tracks_path,
            params={"fields": self.TRACK_FIELDS, "market": "from_token"},
            ignore_expiry=playlist.status_unchanged,
        )

        more_tracks = []
        for page in track_pages:
            if "items" not in page:
                return {}
            more_tracks += page.get("items", [])
        if more_tracks:
            # Take a copy to avoid changing the cached response.
            playlist = copy.deepcopy(playlist)
            playlist.setdefault("tracks", {}).setdefault("items", [])
            playlist["tracks"]["items"] += more_tracks

        return playlist

    def clear_cache(self, extra_expiry=None):
        self._cache.clear()


@unique
class LinkType(Enum):
    TRACK = "track"
    ALBUM = "album"
    ARTIST = "artist"
    PLAYLIST = "playlist"


@dataclass
class WebLink:
    uri: str
    type: LinkType
    id: Optional[str] = None
    owner: Optional[str] = None

    @classmethod
    def from_uri(cls, uri):
        parsed_uri = urllib.parse.urlparse(uri)

        schemes = ("http", "https")
        netlocs = ("open.spotify.com", "play.spotify.com")

        if parsed_uri.scheme == "spotify":
            parts = parsed_uri.path.split(":")
        elif parsed_uri.scheme in schemes and parsed_uri.netloc in netlocs:
            parts = parsed_uri.path[1:].split("/")
        else:
            parts = []

        # Strip out empty parts to ensure we are strict about URI parsing.
        parts = [p for p in parts if p.strip()]

        if len(parts) == 2 and parts[0] in ("track", "album", "artist", "playlist",):
            return cls(uri, LinkType(parts[0]), parts[1], None)
        elif len(parts) == 3 and parts[0] == "user" and parts[2] == "starred":
            if parsed_uri.scheme == "spotify":
                return cls(uri, LinkType.PLAYLIST, None, parts[1])
        elif len(parts) == 3 and parts[0] == "playlist":
            return cls(uri, LinkType.PLAYLIST, parts[2], parts[1])
        elif len(parts) == 4 and parts[0] == "user" and parts[2] == "playlist":
            return cls(uri, LinkType.PLAYLIST, parts[3], parts[1])

        raise ValueError(f"Could not parse {uri!r} as a Spotify URI")



"""This module handles all controls that change the playback."""

from __future__ import annotations

import re
import subprocess
from functools import wraps
from typing import (
    Callable,
    TYPE_CHECKING,
)

from django.core.handlers.wsgi import WSGIRequest
from django.db import transaction
from django.db.models import F
from django.http import HttpResponseForbidden
from django.http.response import HttpResponse, HttpResponseBadRequest

import core.models as models
from core.models import Setting

if TYPE_CHECKING:
    from core.musiq.musiq import Musiq
    from core.musiq.playback import Playback


def control(func: Callable) -> Callable:
    """A decorator for functions that control the playback.
    Every control changes the views state and returns an empty response."""

    def _decorator(
        self: "Controller", request: WSGIRequest, *args, **kwargs
    ) -> HttpResponse:
        # don't allow controls during alarm
        if self.playback.alarm_playing.is_set():
            return HttpResponseBadRequest()
        func(self, request, *args, **kwargs)
        self.musiq.update_state()
        return HttpResponse()

    return wraps(func)(_decorator)


def disabled_when_voting(func: Callable) -> Callable:
    """A decorator for controls that are disabled during voting.
    Only users with appropriate privileges are still able to perform this action."""

    def _decorator(
        self: "Controller", request: WSGIRequest, *args, **kwargs
    ) -> HttpResponse:
        if (
            self.musiq.base.settings.basic.voting_system
            and not self.musiq.base.user_manager.has_controls(request.user)
        ):
            return HttpResponseForbidden()
        func(self, request, *args, **kwargs)
        self.musiq.update_state()
        return HttpResponse()

    return wraps(func)(_decorator)


class Controller:
    """This class provides endpoints for all playback controls."""

    SEEK_DISTANCE = 10 * 1000

    def __init__(self, musiq: "Musiq") -> None:
        self.musiq = musiq
        self.playback: "Playback" = self.musiq.playback

        self.shuffle: bool = (
            self.musiq.base.settings.get_setting("shuffle", "False") == "True"
        )
        self.repeat: bool = self.musiq.base.settings.get_setting(
            "repeat", "False"
        ) == "True"
        self.autoplay: bool = (
            self.musiq.base.settings.get_setting("autoplay", "False") == "True"
        )

    def start(self) -> None:
        try:
            # Try to get the volume from the pulse server.
            active_sink = False
            volume = 100
            for line in subprocess.check_output(
                "pactl list sinks".split(),
                env={"PULSE_SERVER": "127.0.0.1"},
                universal_newlines=True,
            ).splitlines():
                if active_sink and "Volume:" in line:
                    match = re.search(r"(\d+)%", line)
                    if not match:
                        raise ValueError
                    volume = int(match.groups()[0])
                    break
                if "State: RUNNING" in line:
                    active_sink = True
            self.volume = volume / 100
        except (FileNotFoundError, subprocess.CalledProcessError, ValueError):
            with self.playback.mopidy_command(important=True):
                # pulse is not installed or there is no server running.
                # get volume from mopidy
                self.volume = self.playback.player.mixer.get_volume() / 100

    @disabled_when_voting
    @control
    def restart(self, _request: WSGIRequest) -> None:
        """Restarts the current song from the beginning."""
        with self.playback.mopidy_command() as allowed:
            if allowed:
                self.playback.player.playback.seek(0)

    @disabled_when_voting
    @control
    def seek_backward(self, _request: WSGIRequest) -> None:
        """Jumps back in the current song."""
        with self.playback.mopidy_command() as allowed:
            if allowed:
                current_position = self.playback.player.playback.get_time_position()
                self.playback.player.playback.seek(
                    current_position - self.SEEK_DISTANCE
                )

    @disabled_when_voting
    @control
    def play(self, _request: WSGIRequest) -> None:
        """Resumes the current song if it is paused.
        No-op if already playing."""
        with self.playback.mopidy_command() as allowed:
            if allowed:
                self.playback.player.playback.play()

    @disabled_when_voting
    @control
    def pause(self, _request: WSGIRequest) -> None:
        """Pauses the current song if it is playing.
        No-op if already paused."""
        with self.playback.mopidy_command() as allowed:
            if allowed:
                self.playback.player.playback.pause()

    @disabled_when_voting
    @control
    def seek_forward(self, _request: WSGIRequest) -> None:
        """Jumps forward in the current song."""
        with self.playback.mopidy_command() as allowed:
            if allowed:
                current_position = self.playback.player.playback.get_time_position()
                self.playback.player.playback.seek(
                    current_position + self.SEEK_DISTANCE
                )

    @disabled_when_voting
    @control
    def skip(self, _request: WSGIRequest) -> None:
        """Skips the current song and continues with the next one."""
        with self.playback.mopidy_command() as allowed:
            if allowed:
                self.playback.player.playback.next()

    @disabled_when_voting
    @control
    def set_shuffle(self, request: WSGIRequest) -> None:
        """Enables or disables shuffle depending on the given value.
        If enabled, a random song in the queue is chosen as the next one.
        If not, the first one is chosen."""
        enabled = request.POST.get("value") == "true"
        Setting.objects.filter(key="shuffle").update(value=enabled)
        self.shuffle = enabled

    @disabled_when_voting
    @control
    def set_repeat(self, request: WSGIRequest) -> None:
        """Enables or disables repeat depending on the given value.
        If enabled, a song is enqueued again after it finished playing."""
        enabled = request.POST.get("value") == "true"
        Setting.objects.filter(key="repeat").update(value=enabled)
        self.repeat = enabled

    @disabled_when_voting
    @control
    def set_autoplay(self, request: WSGIRequest) -> None:
        """Enables or disables autoplay depending on the given value.
        If enabled and the current song is the last one,
        a new song is enqueued, based on the current one."""
        enabled = request.POST.get("value") == "true"
        Setting.objects.filter(key="autoplay").update(value=enabled)
        self.autoplay = enabled
        self.playback.handle_autoplay()

    @disabled_when_voting
    @control
    def set_volume(self, request: WSGIRequest) -> None:
        """Sets the playback volume.
        value has to be a float between 0 and 1."""
        self.volume = float(request.POST.get("value"))  # type: ignore
        try:
            # Try to set the volume via the pulse server.
            # This is faster and does not impact visualization
            subprocess.run(
                f"pactl set-sink-volume @DEFAULT_SINK@ {round(self.volume*100)}%".split(),
                env={"PULSE_SERVER": "127.0.0.1"},
                check=True,
            )
        except (FileNotFoundError, subprocess.CalledProcessError):
            # pulse is not installed or there is no server running.
            # change mopidy's volume
            with self.playback.mopidy_command() as allowed:
                if allowed:
                    self.playback.player.mixer.set_volume(round(self.volume * 100))

    @disabled_when_voting
    @control
    def remove_all(self, request: WSGIRequest) -> HttpResponse:
        """Empties the queue. Only admin is permitted to do this."""
        if not self.musiq.base.user_manager.is_admin(request.user):
            return HttpResponseForbidden()
        with self.playback.mopidy_command() as allowed:
            if allowed:
                with transaction.atomic():
                    count = self.playback.queue.count()
                    self.playback.queue.all().delete()
                for _ in range(count):
                    self.playback.queue_semaphore.acquire(blocking=False)
        return HttpResponse()

    @disabled_when_voting
    @control
    def prioritize(self, request: WSGIRequest) -> HttpResponse:
        """Prioritizes song by making it the first one in the queue."""
        key = request.POST.get("key")
        if key is None:
            return HttpResponseBadRequest()
        ikey = int(key)
        self.playback.queue.prioritize(ikey)
        return HttpResponse()

    @disabled_when_voting
    @control
    def remove(self, request: WSGIRequest) -> HttpResponse:
        """Removes a song identified by the given key from the queue."""
        key = request.POST.get("key")
        if key is None:
            return HttpResponseBadRequest()
        ikey = int(key)
        try:
            removed = self.playback.queue.remove(ikey)
            self.playback.queue_semaphore.acquire(blocking=False)
            # if we removed a song and it was added by autoplay,
            # we want it to be the new basis for autoplay
            if not removed.manually_requested:
                self.playback.handle_autoplay(removed.external_url or removed.title)
            else:
                self.playback.handle_autoplay()
        except models.QueuedSong.DoesNotExist:
            return HttpResponseBadRequest("song does not exist")
        return HttpResponse()

    @disabled_when_voting
    @control
    def reorder(self, request: WSGIRequest) -> HttpResponse:
        """Reorders the queue.
        The song specified by element is inserted between prev and next."""
        prev_key = request.POST.get("prev")
        cur_key = request.POST.get("element")
        next_key = request.POST.get("next")
        if not cur_key:
            return HttpResponseBadRequest()
        if not prev_key:
            iprev_key = None
        else:
            iprev_key = int(prev_key)
        icur_key = int(cur_key)
        if not next_key:
            inext_key = None
        else:
            inext_key = int(next_key)
        try:
            self.playback.queue.reorder(iprev_key, icur_key, inext_key)
        except ValueError:
            return HttpResponseBadRequest("request on old state")
        return HttpResponse()

    @control
    def vote_up(self, request: WSGIRequest) -> HttpResponse:
        """Increases the vote-count of the given song by one."""
        key = request.POST.get("key")
        if key is None:
            return HttpResponseBadRequest()
        ikey = int(key)

        models.CurrentSong.objects.filter(queue_key=ikey).update(votes=F("votes") + 1)
        self.playback.queue.vote_up(ikey)
        return HttpResponse()

    @control
    def vote_down(self, request: WSGIRequest) -> HttpResponse:
        """Decreases the vote-count of the given song by one.
        If a song receives too many downvotes, it is removed."""
        key = request.POST.get("key")
        if key is None:
            return HttpResponseBadRequest()
        ikey = int(key)

        models.CurrentSong.objects.filter(queue_key=ikey).update(votes=F("votes") - 1)
        try:
            current_song = models.CurrentSong.objects.get()
            if (
                current_song.queue_key == ikey
                and current_song.votes
                <= -self.musiq.base.settings.basic.downvotes_to_kick
            ):
                with self.playback.mopidy_command() as allowed:
                    if allowed:
                        self.playback.player.playback.next()
        except models.CurrentSong.DoesNotExist:
            pass

        removed = self.playback.queue.vote_down(
            ikey, -self.musiq.base.settings.basic.downvotes_to_kick
        )
        # if we removed a song by voting, and it was added by autoplay,
        # we want it to be the new basis for autoplay
        if removed is not None:
            self.playback.queue_semaphore.acquire(blocking=False)
            if not removed.manually_requested:
                self.playback.handle_autoplay(removed.external_url or removed.title)
            else:
                self.playback.handle_autoplay()
        return HttpResponse()



# Generated by Django 2.2.13 on 2020-07-26 20:12

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("core", "0008_auto_20200514_1213"),
    ]

    operations = [
        migrations.DeleteModel(name="Pad",),
    ]



"""This module handles all settings related to the local library."""

from __future__ import annotations

import logging
import os
import time
from typing import TYPE_CHECKING

from django.conf import settings
from django.core.handlers.wsgi import WSGIRequest
from django.http import HttpResponse
from django.http import HttpResponseBadRequest
from django.http import JsonResponse
from mutagen import MutagenError

import core.musiq.song_utils as song_utils
from core.models import ArchivedSong, ArchivedPlaylist, PlaylistEntry
from core.settings.settings import Settings
from core.util import background_thread

if TYPE_CHECKING:
    from core.base import Base


class Library:
    """This class is responsible for handling settings changes related to the local library."""

    def __init__(self, base: "Base"):
        self.base = base
        self.scan_progress = "0 / 0 / 0"

    @staticmethod
    def get_library_path() -> str:
        return os.path.abspath(os.path.join(settings.SONGS_CACHE_DIR, "local_library"))

    @Settings.option
    def list_subdirectories(self, request: WSGIRequest) -> HttpResponse:
        """Returns a list of all subdirectories for the given path."""
        path = request.GET.get("path")
        if path is None:
            return HttpResponseBadRequest("path was not supplied.")
        basedir, subdirpart = os.path.split(path)
        if path == "":
            suggestions = ["/"]
        elif os.path.isdir(basedir):
            suggestions = [
                os.path.join(basedir, subdir + "/")
                for subdir in next(os.walk(basedir))[1]
                if subdir.lower().startswith(subdirpart.lower())
            ]
            suggestions.sort()
        else:
            suggestions = ["not a valid directory"]
        if not suggestions:
            suggestions = ["not a valid directory"]
        return JsonResponse(suggestions, safe=False)

    @Settings.option
    def scan_library(self, request: WSGIRequest) -> HttpResponse:
        """Scan the folder at the given path and add all its sound files to the database."""
        library_path = request.POST.get("library_path")
        if library_path is None:
            return HttpResponseBadRequest("library path was not supplied.")

        if not os.path.isdir(library_path):
            return HttpResponseBadRequest("not a directory")
        library_path = os.path.abspath(library_path)

        self.scan_progress = "0 / 0 / 0"
        self.base.settings.update_state()

        self._scan_library(library_path)

        return HttpResponse(
            f"started scanning in {library_path}. This could take a while"
        )

    @background_thread
    def _scan_library(self, library_path: str) -> None:
        scan_start = time.time()
        last_update = scan_start
        update_frequency = 0.5
        filecount = 0
        for (dirpath, _, filenames) in os.walk(library_path):
            now = time.time()
            if now - last_update > update_frequency:
                last_update = now
                self.scan_progress = f"{filecount} / 0 / 0"
                self.base.settings.update_state()
            if os.path.abspath(dirpath) == os.path.abspath(settings.SONGS_CACHE_DIR):
                # do not add files handled by raveberry as local files
                continue
            filecount += len(filenames)

        library_link = os.path.join(settings.SONGS_CACHE_DIR, "local_library")
        try:
            os.remove(library_link)
        except FileNotFoundError:
            pass
        os.symlink(library_path, library_link)

        logging.info("started scanning in %s", library_path)

        self.scan_progress = f"{filecount} / 0 / 0"
        self.base.settings.update_state()

        files_scanned = 0
        files_added = 0
        for (dirpath, _, filenames) in os.walk(library_path):
            if os.path.abspath(dirpath) == os.path.abspath(settings.SONGS_CACHE_DIR):
                # do not add files handled by raveberry as local files
                continue
            now = time.time()
            if now - last_update > update_frequency:
                last_update = now
                self.scan_progress = f"{filecount} / {files_scanned} / {files_added}"
                self.base.settings.update_state()
            for filename in filenames:
                files_scanned += 1
                path = os.path.join(dirpath, filename)
                try:
                    metadata = song_utils.get_metadata(path)
                except (ValueError, MutagenError):
                    # the given file could not be parsed and will not be added to the database
                    pass
                else:
                    library_relative_path = path[len(library_path) + 1 :]
                    external_url = os.path.join("local_library", library_relative_path)
                    if not ArchivedSong.objects.filter(url=external_url).exists():
                        files_added += 1
                        ArchivedSong.objects.create(
                            url=external_url,
                            artist=metadata["artist"],
                            title=metadata["title"],
                            counter=0,
                        )

        assert files_scanned == filecount
        self.scan_progress = f"{filecount} / {files_scanned} / {files_added}"
        self.base.settings.update_state()

        logging.info("done scanning in %s", library_path)

    @Settings.option
    def create_playlists(self, _request: WSGIRequest) -> HttpResponse:
        """Create a playlist for every folder in the library."""
        library_link = os.path.join(settings.SONGS_CACHE_DIR, "local_library")
        if not os.path.islink(library_link):
            return HttpResponseBadRequest("No library set")

        self.scan_progress = f"0 / 0 / 0"
        self.base.settings.update_state()

        self._create_playlists()

        return HttpResponse(f"started creating playlsts. This could take a while")

    @background_thread
    def _create_playlists(self) -> None:
        local_files = ArchivedSong.objects.filter(
            url__startswith="local_library"
        ).count()

        library_link = os.path.join(settings.SONGS_CACHE_DIR, "local_library")
        library_path = os.path.abspath(library_link)

        logging.info("started creating playlists in %s", library_path)

        self.scan_progress = f"{local_files} / 0 / 0"
        self.base.settings.update_state()

        scan_start = time.time()
        last_update = scan_start
        update_frequency = 0.5
        files_processed = 0
        files_added = 0
        for (dirpath, _, filenames) in os.walk(library_path):
            now = time.time()
            if now - last_update > update_frequency:
                last_update = now
                self.scan_progress = (
                    f"{local_files} / {files_processed} / {files_added}"
                )
                self.base.settings.update_state()

            song_urls = []
            # unfortunately there is no way to access track numbers accross different file types
            # so we have to add songs to playlists alphabetically
            for filename in sorted(filenames):
                path = os.path.join(dirpath, filename)
                library_relative_path = path[len(library_path) + 1 :]
                external_url = os.path.join("local_library", library_relative_path)
                if ArchivedSong.objects.filter(url=external_url).exists():
                    files_processed += 1
                    song_urls.append(external_url)

            if not song_urls:
                continue

            playlist_id = os.path.join(
                "local_library", dirpath[len(library_path) + 1 :]
            )
            playlist_title = os.path.split(dirpath)[1]
            playlist, created = ArchivedPlaylist.objects.get_or_create(
                list_id=playlist_id, title=playlist_title, counter=0
            )
            if not created:
                # this playlist already exists, skip
                continue

            song_index = 0
            for external_url in song_urls:
                PlaylistEntry.objects.create(
                    playlist=playlist, index=song_index, url=external_url,
                )
                files_added += 1
                song_index += 1

        self.scan_progress = f"{local_files} / {files_processed} / {files_added}"
        self.base.settings.update_state()



"""ASGI entrypoint. Configures Django and then runs the application
defined in the ASGI_APPLICATION setting."""

import os
import django
from channels.routing import get_default_application

os.environ.setdefault("DJANGO_SETTINGS_MODULE", "main.settings")
django.setup()
application = get_default_application()



"""This module handles all settings related to wifi."""

from __future__ import annotations

import os
import subprocess
from typing import TYPE_CHECKING

from django.conf import settings
from django.core.handlers.wsgi import WSGIRequest
from django.http import HttpResponse
from django.http import HttpResponseBadRequest
from django.http import JsonResponse

from core.settings.settings import Settings

if TYPE_CHECKING:
    from core.base import Base


class Wifi:
    """This class is responsible for handling settings changes related to wifi."""

    def __init__(self, base: "Base"):
        self.base = base

    @Settings.option
    def available_ssids(self, _request: WSGIRequest) -> JsonResponse:
        """List all ssids that can currently be seen."""
        output = subprocess.check_output(
            ["sudo", "/usr/local/sbin/raveberry/list_available_ssids"]
        ).decode()
        ssids = output.split("\n")
        return JsonResponse(ssids[:-1], safe=False)

    @Settings.option
    def connect_to_wifi(self, request: WSGIRequest) -> HttpResponse:
        """Connect to a given ssid with the given password."""
        ssid = request.POST.get("ssid")
        password = request.POST.get("password")
        if ssid is None or password is None or ssid == "" or password == "":
            return HttpResponseBadRequest("Please provide both SSID and password")
        try:
            output = subprocess.check_output(
                ["sudo", "/usr/local/sbin/raveberry/connect_to_wifi", ssid, password]
            ).decode()
            return HttpResponse(output)
        except subprocess.CalledProcessError as e:
            output = e.output.decode()
            return HttpResponseBadRequest(output)

    @Settings.option
    def disable_homewifi(self, _request: WSGIRequest) -> None:
        """Disable homewifi function."""
        subprocess.call(["sudo", "/usr/local/sbin/raveberry/disable_homewifi"])

    @Settings.option
    def enable_homewifi(self, _request: WSGIRequest) -> None:
        """Enable homewifi function."""
        subprocess.call(["sudo", "/usr/local/sbin/raveberry/enable_homewifi"])

    @Settings.option
    def stored_ssids(self, _request: WSGIRequest) -> JsonResponse:
        """Return the list of ssids that this Raspberry Pi was connected to in the past."""
        output = subprocess.check_output(
            ["sudo", "/usr/local/sbin/raveberry/list_stored_ssids"]
        ).decode()
        ssids = output.split("\n")
        return JsonResponse(ssids[:-1], safe=False)

    @Settings.option
    def set_homewifi_ssid(self, request: WSGIRequest) -> HttpResponse:
        """Set the home network.
        The hotspot will not be created if connected to this wifi."""
        homewifi_ssid = request.POST.get("homewifi_ssid")
        if homewifi_ssid is None:
            return HttpResponseBadRequest("homewifi ssid was not supplied.")
        with open(os.path.join(settings.BASE_DIR, "config/homewifi"), "w+") as f:
            f.write(homewifi_ssid)
        return HttpResponse()



"""This module contains all url endpoints and maps them to their corresponding functions."""

import re

from django.urls import include, URLPattern
from django.urls import path
from django.views.generic import RedirectView

from core import mock
from main import settings

from core.base import Base

BASE = Base()

MUSIQ_URLS = [
    path("state/", BASE.musiq.get_state, name="musiq_state"),
    path(
        "random_suggestion/",
        BASE.musiq.suggestions.random_suggestion,
        name="random_suggestion",
    ),
    path("request_music/", BASE.musiq.request_music, name="request_music"),
    path("suggestions/", BASE.musiq.suggestions.get_suggestions, name="suggestions"),
    path("restart/", BASE.musiq.controller.restart, name="restart_song"),
    path("seek_backward/", BASE.musiq.controller.seek_backward, name="seek_backward"),
    path("play/", BASE.musiq.controller.play, name="play_song"),
    path("pause/", BASE.musiq.controller.pause, name="pause_song"),
    path("seek_forward/", BASE.musiq.controller.seek_forward, name="seek_forward"),
    path("skip/", BASE.musiq.controller.skip, name="skip_song"),
    path("set_shuffle/", BASE.musiq.controller.set_shuffle, name="set_shuffle"),
    path("set_repeat/", BASE.musiq.controller.set_repeat, name="set_repeat"),
    path("set_autoplay/", BASE.musiq.controller.set_autoplay, name="set_autoplay"),
    path("request_radio/", BASE.musiq.request_radio, name="request_radio"),
    path("set_volume/", BASE.musiq.controller.set_volume, name="set_volume"),
    path("remove_all/", BASE.musiq.controller.remove_all, name="remove_all"),
    path("prioritize/", BASE.musiq.controller.prioritize, name="prioritize_song"),
    path("remove/", BASE.musiq.controller.remove, name="remove_song"),
    path("reorder/", BASE.musiq.controller.reorder, name="reorder_song"),
    path("vote_up/", BASE.musiq.controller.vote_up, name="vote_up_song"),
    path("vote_down/", BASE.musiq.controller.vote_down, name="vote_down_song"),
]

LIGHTS_URLS = [
    path("state/", BASE.lights.get_state, name="lights_state"),
    path(
        "shortcut/",
        BASE.lights.controller.set_lights_shortcut,
        name="set_lights_shortcut",
    ),
    path(
        "set_ring_program/",
        BASE.lights.controller.set_ring_program,
        name="set_ring_program",
    ),
    path(
        "set_ring_brightness/",
        BASE.lights.controller.set_ring_brightness,
        name="set_ring_brightness",
    ),
    path(
        "set_ring_monochrome/",
        BASE.lights.controller.set_ring_monochrome,
        name="set_ring_monochrome",
    ),
    path(
        "set_wled_led_count/",
        BASE.lights.controller.set_wled_led_count,
        name="set_wled_led_count",
    ),
    path("set_wled_ip/", BASE.lights.controller.set_wled_ip, name="set_wled_ip",),
    path("set_wled_port/", BASE.lights.controller.set_wled_port, name="set_wled_port",),
    path(
        "set_wled_program/",
        BASE.lights.controller.set_wled_program,
        name="set_wled_program",
    ),
    path(
        "set_wled_brightness/",
        BASE.lights.controller.set_wled_brightness,
        name="set_wled_brightness",
    ),
    path(
        "set_wled_monochrome/",
        BASE.lights.controller.set_wled_monochrome,
        name="set_wled_monochrome",
    ),
    path(
        "set_strip_program/",
        BASE.lights.controller.set_strip_program,
        name="set_strip_program",
    ),
    path(
        "set_strip_brightness/",
        BASE.lights.controller.set_strip_brightness,
        name="set_strip_brightness",
    ),
    path("adjust_screen/", BASE.lights.controller.adjust_screen, name="adjust_screen"),
    path(
        "set_screen_program/",
        BASE.lights.controller.set_screen_program,
        name="set_screen_program",
    ),
    path(
        "set_program_speed/",
        BASE.lights.controller.set_program_speed,
        name="set_program_speed",
    ),
    path(
        "set_fixed_color/",
        BASE.lights.controller.set_fixed_color,
        name="set_fixed_color",
    ),
]

SETTINGS_URLS = [
    path("state/", BASE.settings.get_state, name="settings_state"),
    path(
        "set_voting_system/",
        BASE.settings.basic.set_voting_system,
        name="set_voting_system",
    ),
    path(
        "set_new_music_only/",
        BASE.settings.basic.set_new_music_only,
        name="set_new_music_only",
    ),
    path(
        "set_logging_enabled/",
        BASE.settings.basic.set_logging_enabled,
        name="set_logging_enabled",
    ),
    path(
        "set_online_suggestions/",
        BASE.settings.basic.set_online_suggestions,
        name="set_online_suggestions",
    ),
    path(
        "set_number_of_suggestions/",
        BASE.settings.basic.set_number_of_suggestions,
        name="set_number_of_suggestions",
    ),
    path(
        "set_people_to_party/",
        BASE.settings.basic.set_people_to_party,
        name="set_people_to_party",
    ),
    path(
        "set_alarm_probability/",
        BASE.settings.basic.set_alarm_probability,
        name="set_alarm_probability",
    ),
    path(
        "set_downvotes_to_kick/",
        BASE.settings.basic.set_downvotes_to_kick,
        name="set_downvotes_to_kick",
    ),
    path(
        "set_max_download_size/",
        BASE.settings.basic.set_max_download_size,
        name="set_max_download_size",
    ),
    path(
        "set_max_playlist_items/",
        BASE.settings.basic.set_max_playlist_items,
        name="set_max_playlist_items",
    ),
    path("check_internet/", BASE.settings.basic.check_internet, name="check_internet"),
    path(
        "update_user_count/",
        BASE.settings.basic.update_user_count,
        name="update_user_count",
    ),
    path(
        "set_youtube_enabled/",
        BASE.settings.platforms.set_youtube_enabled,
        name="set_youtube_enabled",
    ),
    path(
        "set_youtube_suggestions/",
        BASE.settings.platforms.set_youtube_suggestions,
        name="set_youtube_suggestions",
    ),
    path(
        "set_spotify_enabled/",
        BASE.settings.platforms.set_spotify_enabled,
        name="set_spotify_enabled",
    ),
    path(
        "set_spotify_suggestions/",
        BASE.settings.platforms.set_spotify_suggestions,
        name="set_spotify_suggestions",
    ),
    path(
        "set_spotify_credentials/",
        BASE.settings.platforms.set_spotify_credentials,
        name="set_spotify_credentials",
    ),
    path(
        "set_soundcloud_enabled/",
        BASE.settings.platforms.set_soundcloud_enabled,
        name="set_soundcloud_enabled",
    ),
    path(
        "set_soundcloud_suggestions/",
        BASE.settings.platforms.set_soundcloud_suggestions,
        name="set_soundcloud_suggestions",
    ),
    path(
        "set_soundcloud_credentials/",
        BASE.settings.platforms.set_soundcloud_credentials,
        name="set_soundcloud_credentials",
    ),
    path(
        "set_bluetooth_scanning/",
        BASE.settings.sound.set_bluetooth_scanning,
        name="set_bluetooth_scanning",
    ),
    path(
        "connect_bluetooth/",
        BASE.settings.sound.connect_bluetooth,
        name="connect_bluetooth",
    ),
    path(
        "disconnect_bluetooth/",
        BASE.settings.sound.disconnect_bluetooth,
        name="disconnect_bluetooth",
    ),
    path("output_devices/", BASE.settings.sound.output_devices, name="output_devices",),
    path(
        "set_output_device/",
        BASE.settings.sound.set_output_device,
        name="set_output_device",
    ),
    path(
        "available_ssids/", BASE.settings.wifi.available_ssids, name="available_ssids",
    ),
    path(
        "connect_to_wifi/", BASE.settings.wifi.connect_to_wifi, name="connect_to_wifi",
    ),
    path(
        "enable_homewifi/", BASE.settings.wifi.enable_homewifi, name="enable_homewifi",
    ),
    path(
        "disable_homewifi/",
        BASE.settings.wifi.disable_homewifi,
        name="disable_homewifi",
    ),
    path("stored_ssids/", BASE.settings.wifi.stored_ssids, name="stored_ssids"),
    path(
        "set_homewifi_ssid/",
        BASE.settings.wifi.set_homewifi_ssid,
        name="set_homewifi_ssid",
    ),
    path(
        "list_subdirectories/",
        BASE.settings.library.list_subdirectories,
        name="list_subdirectories",
    ),
    path("scan_library/", BASE.settings.library.scan_library, name="scan_library"),
    path(
        "create_playlists/",
        BASE.settings.library.create_playlists,
        name="create_playlists",
    ),
    path("analyse/", BASE.settings.analysis.analyse, name="analyse"),
    path(
        "disable_streaming/",
        BASE.settings.system.disable_streaming,
        name="disable_streaming",
    ),
    path(
        "enable_streaming/",
        BASE.settings.system.enable_streaming,
        name="enable_streaming",
    ),
    path(
        "disable_events/", BASE.settings.system.disable_events, name="disable_events",
    ),
    path("enable_events/", BASE.settings.system.enable_events, name="enable_events"),
    path(
        "disable_hotspot/",
        BASE.settings.system.disable_hotspot,
        name="disable_hotspot",
    ),
    path(
        "enable_hotspot/", BASE.settings.system.enable_hotspot, name="enable_hotspot",
    ),
    path(
        "unprotect_wifi/", BASE.settings.system.unprotect_wifi, name="unprotect_wifi",
    ),
    path("protect_wifi/", BASE.settings.system.protect_wifi, name="protect_wifi"),
    path(
        "disable_tunneling/",
        BASE.settings.system.disable_tunneling,
        name="disable_tunneling",
    ),
    path(
        "enable_tunneling/",
        BASE.settings.system.enable_tunneling,
        name="enable_tunneling",
    ),
    path(
        "disable_remote/", BASE.settings.system.disable_remote, name="disable_remote",
    ),
    path("enable_remote/", BASE.settings.system.enable_remote, name="enable_remote"),
    path("reboot_server/", BASE.settings.system.reboot_server, name="reboot_server"),
    path("reboot_system/", BASE.settings.system.reboot_system, name="reboot_system"),
    path(
        "shutdown_system/", BASE.settings.system.shutdown_system, name="shutdown_system"
    ),
    path(
        "get_latest_version/",
        BASE.settings.system.get_latest_version,
        name="get_latest_version",
    ),
    path(
        "get_upgrade_config/",
        BASE.settings.system.get_upgrade_config,
        name="get_upgrade_config",
    ),
    path(
        "upgrade_raveberry/",
        BASE.settings.system.upgrade_raveberry,
        name="upgrade_raveberry",
    ),
]

urlpatterns = [
    path("", RedirectView.as_view(pattern_name="musiq", permanent=False), name="base"),
    path("musiq/", BASE.musiq.index, name="musiq"),
    path("lights/", BASE.lights.index, name="lights"),
    path("stream/", BASE.no_stream, name="no_stream"),
    path("network_info/", BASE.network_info.index, name="network_info"),
    path("settings/", BASE.settings.index, name="settings"),
    path("accounts/", include("django.contrib.auth.urls")),
    path("login/", RedirectView.as_view(pattern_name="login", permanent=False)),
    path("logged_in/", BASE.logged_in, name="logged_in"),
    path("logout/", RedirectView.as_view(pattern_name="logout", permanent=False)),
    path(
        "ajax/",
        include(
            [
                path("state/", BASE.get_state, name="base_state"),
                path("submit_hashtag/", BASE.submit_hashtag, name="submit_hashtag"),
                path("musiq/", include(MUSIQ_URLS)),
                path("lights/", include(LIGHTS_URLS)),
                path("settings/", include(SETTINGS_URLS)),
            ]
        ),
    ),
    path(
        "api/",
        include(
            [
                path(
                    "musiq/",
                    include(
                        [path("post_song/", BASE.musiq.post_song, name="post_song"),]
                    ),
                ),
            ]
        ),
    ),
]

if settings.MOCK:
    for url in urlpatterns:
        if isinstance(url, URLPattern):
            url.callback = mock.index
else:
    BASE.start()



"""This module contains the device superclass."""


class Device:
    """A class representing a visualization device that Raveberry can control."""

    def __init__(self, lights, name) -> None:
        self.lights = lights
        self.name = name
        self.brightness = 1.0
        self.monochrome = False
        self.initialized = False
        self.last_program = self.lights.disabled_program
        self.program = self.lights.disabled_program

    def load_program(self) -> None:
        last_program_name = self.lights.base.settings.get_setting(
            f"last_{self.name}_program", "Disabled"
        )
        program_name = self.lights.base.settings.get_setting(
            f"{self.name}_program", "Disabled"
        )

        self.last_program = self.lights.all_programs[last_program_name]
        # only enable if the device is initialized
        if self.initialized:
            self.program = self.lights.all_programs[program_name]

        self.program.use()

    def clear(self) -> None:
        """Resets this device, clearing all visualization."""
        raise NotImplementedError()
